import 'delegating_array.dart';
import 'package:meta/meta.dart';
import 'array.dart';
import 'dart:ffi';

/// Reference counted Array. Not thread safe!
class SharedArray<T extends NativeType> extends DelegatingArray<T> {
  int _counter = 0;
  @nonVirtual
  int get counter => _counter;
  @nonVirtual
  bool get deleted => counter == -1;

  /// Takes ownership of [arr]
  SharedArray(Array<T> arr) : super(arr);

  /// Returns [this] and updates [counter]. Throws [StateError]
  /// when [this] has already been deleted or when [counter] overflows.
  SharedArray<T> newShared() {
    if (deleted) {
      throw StateError(
          'Cannot create new new SharedArray from deleted SharedArray');
    }
    _counter++;
    if (counter.isNegative) {
      throw StateError(
          'Cannot create any more SharedArray. Counter reached maximum integer value');
    }
    return this;
  }

  /// Deletes the shared [Array], even with `[counter] > 0`. Calling this method multiple times does nothing.
  @mustCallSuper
  void forceDelete() {
    if (deleted) return;
    _counter = -1;
    super.delete();
  }

  /// Deletes [_array] when [counter] has reached `0`. Calling this method multiple times
  /// does nothing.
  @override
  @mustCallSuper
  void delete() {
    if (deleted) return;
    _counter--;
    if (counter == 0) {
      super.delete();
      _counter = -1;
    }
  }
}
